import { Activity } from '../Models/Activity';
import { Matrix } from '../Models/Matrix';
import { IAccessory } from './IAccessory';
import callbackify from '../Util/Callbackify';
import HarmonyDataProvider from '../DataProviders/HarmonyDataProvider';

let Service: HAPNodeJS.Service;
let Characteristic: HAPNodeJS.Characteristic;
let Api: any;

export interface IControlUnitProps {
    dataProvider: HarmonyDataProvider,
    displayName: string,
    activities: Array<Activity>,
    api: any,
    log: any,
}

/**
 * ControlUnit accessory
 */
export class ControlUnit implements IAccessory {
    //Required by homebridge
    name: string = "";

    //fields
    private log: any = {};
    private displayName: string = "";

    //Service fields
    private televisionService: HAPNodeJS.Service | undefined;
    private televisionSpeakerService: HAPNodeJS.Service | undefined;
    private informationService: HAPNodeJS.Service | undefined;
    private inputServices: Array<HAPNodeJS.Service | undefined> = [];

    //Harmony fields
    private activities: Array<Activity> = [];
    private dataProvider: HarmonyDataProvider;

    /**
     * Constructor
     * @param props Input properties
     */
    constructor(props: IControlUnitProps) {
        //Assign class variables
        this.log = props.log;
        Api = props.api;
        Service = props.api.hap.Service;
        Characteristic = props.api.hap.Characteristic;
        this.displayName = props.displayName;
        this.name = this.displayName;

        this.activities = props.activities;

        this.dataProvider = props.dataProvider;

        //Configure services
        this.configureTvService();
        this.configureTvSpeakerService();
        // this.configureAccessoryInformation();
        this.configureInputSourceService();
    }
    /*************
     * 
     * Tv Service
     * 
     *************/

    /**
     * Configure television service
     */
    private configureTvService(): void {
        this.televisionService = new Service.Television(
            this.displayName,
            'tvService'
        )

        this.televisionService.setCharacteristic(Characteristic.ConfiguredName, this.displayName);
        //@ts-ignore
        this.televisionService.setCharacteristic(Characteristic.SleepDiscoveryMode, Characteristic.SleepDiscoveryMode.ALWAYS_DISCOVERABLE);
        this.televisionService.setCharacteristic(Characteristic.ActiveIdentifier, 1);
        this.televisionService.setCharacteristic(Characteristic.Active, false);

        //setup listeners
        this.televisionService.getCharacteristic(Characteristic.Active)
            //@ts-ignore
            .on("set", callbackify(this.onSetAccessoryActive))
            //@ts-ignore
            .on("get", callbackify(this.onGetAccessoryActive));

        this.televisionService.getCharacteristic(Characteristic.RemoteKey)
            //@ts-ignore
            .on("set", callbackify(this.onSetRemoteKey));

        this.televisionService.getCharacteristic(Characteristic.ActiveIdentifier)
            //@ts-ignore
            .on("set", callbackify(this.onSetActiveIdentifier))
            //@ts-ignore
            .on("get", callbackify(this.onGetActiveIdentifier));
    }

    //TODO
    private onSetAccessoryActive = async (value: any) => {
        this.log(`set active + ${value}`);
        switch (value) {
            case 0: this.dataProvider.powerOff(this.name); break;
            case 1: this.dataProvider.powerOn(this.name, this.activities[0]); break;
        }
    }

    //TODO
    private onGetAccessoryActive = async () => {
        this.log(`get active`)

        //@ts-ignore
        return this.dataProvider.getIsActive ? Characteristic.Active.Active : Characteristic.Active.Inactive
        // return Characteristic.Active.Active;
    }

    //TODO
    private onSetRemoteKey = async (key: any) => {
        this.log(`set remote key + ${key}`);
    }

    private onSetActiveIdentifier = async (identifier: any) => {
        this.log(`set active identifier + ${identifier}`);
    }

    private onGetActiveIdentifier = async () => {
        this.log(`get active identifier`);
        return 1;
    }

    /******************
     * 
     * Speaker Service
     * 
     *****************/

    /**
     * Configure Speaker Service
     */
    private configureTvSpeakerService(): void {
        this.televisionSpeakerService = new Service.TelevisionSpeaker(
            this.displayName,
            'tvSpeakerService'
        );
        this.televisionSpeakerService.setCharacteristic(Characteristic.Name, this.displayName);
        //@ts-ignore
        this.televisionSpeakerService.setCharacteristic(Characteristic.Active, Characteristic.Active.Active);
        //@ts-ignore
        this.televisionSpeakerService.setCharacteristic(Characteristic.VolumeControlType, Characteristic.VolumeControlType.Absolute);
        this.televisionSpeakerService.subtype = this.displayName + "Volume";
        if (this.televisionService) {
            //@ts-ignore
            this.televisionService.addLinkedService(this.televisionSpeakerService);
        }

        //Setup listeners
        this.televisionSpeakerService.getCharacteristic(Characteristic.VolumeSelector)
            //@ts-ignore
            .on("set", callbackify(this.onSetVolumeSelector));
    }

    //TODO
    private onSetVolumeSelector = async (value: any) => {
        this.log(`set volume + ${value}`);
    }

    /*********************
     * 
     * Information Service
     * 
     ********************/

    /**
     * Configure information service
     */
    private configureAccessoryInformation(): void {
        this.informationService = new Service.AccessoryInformation(this.displayName, 'information');
        this.informationService
            .setCharacteristic(Characteristic.Manufacturer, 'Loftux Carwings')
            .setCharacteristic(Characteristic.Model, 'Heater-Cooler')
    }

    /*****************
     * 
     * Input services
     * 
     *****************/

    /**
     * Configure input service
     */
    private configureInputSourceService(): void {
        let inputs: Array<HAPNodeJS.Service> = [];
        this.activities.forEach((activity: Activity, index: number) => {
            this.log(activity.displayName);
            let inputService = new Service.InputSource(activity.displayName, 'activity' + activity.displayName);
            inputService
                .setCharacteristic(Characteristic.Identifier, index)
                .setCharacteristic(
                    Characteristic.ConfiguredName,
                    activity.displayName)
                .setCharacteristic(
                    Characteristic.IsConfigured,
                    //@ts-ignore
                    Characteristic.IsConfigured.CONFIGURED
                )
                //@ts-ignore
                .setCharacteristic(Characteristic.InputSourceType, Characteristic.InputSourceType.HDMI);

            //@ts-ignore
            this.televisionService!.addLinkedService(inputService);
            inputs.push(inputService);
        });
        this.inputServices = inputs;
    }



    /**
     * Called by homebridge to gather services for this accessory.
     */
    getServices(): Array<HAPNodeJS.Service> {
        this.log(`${this.displayName}`)
        let services: Array<HAPNodeJS.Service> = [this.televisionService!, this.televisionSpeakerService!];
        this.inputServices.forEach((service: HAPNodeJS.Service | undefined) => {
            services.push(service!);
        });
        return (services);
    }
}